
# This code is not particularly well-structured right now, but here is the idea.
# In ipopt_optimize, you can pass in an arbitrary "conderivatives=..." kwarg.
# This should be a struct that has the two methods _con_jacobian and
# _con_lagpart_hess with the signatures shown below, which give back what you
# would think. This is an example of using AD to do this with no closures.

struct AutoFwdConDerivatives{F}
  cons::Constraints{F}
end

struct LagrangianConTerm{F,L} <: Function
  cons::Constraints{F}
  lam::L
end

struct ConstraintJacobian{C} <: Function
  conderivatives::C
  jac_buf::Matrix{Float64}
  con_buf::Vector{Float64}
end

# This is the first method that your custom constraint derivative wrapper needs
# to implement!
function _con_jacobian(jacbuf, conbuf, CD::AutoFwdConDerivatives{F}, x) where{F}
  cons = CD.cons
  ForwardDiff.jacobian!(jacbuf, CD.cons.fun, conbuf, x)
end

function (lt::LagrangianConTerm{F})(x) where{F}
  cons     = lt.cons
  innerbuf = zeros(eltype(x), cons.ncon)
  cons.fun(innerbuf, x)
  dot(view(lt.lam, 1:cons.ncon), innerbuf)
end

# This is the second!
function _con_lagpart_hess(hesbuf, lam, CD::AutoFwdConDerivatives{F}, x) where{F}
  lagrange_con = LagrangianConTerm(CD.cons, lam)
  ForwardDiff.hessian!(hesbuf, lagrange_con, x)
end

function (cj::ConstraintJacobian{F})(p, r, c, v) where{F}
  ncon = cj.conderivatives.cons.ncon
  isnothing(v) && return jac_structure!(r, c, ncon, length(p)) 
  _con_jacobian(cj.jac_buf, cj.con_buf, cj.conderivatives, p)
  v .= vec(permutedims(cj.jac_buf))
end

