
abstract type DerivativeSpec end

struct AutoFwdDerivatives{F,R,C} <: DerivativeSpec
  fn::F
  hess_res::R
  hess_cfg::C
  hhash::Vector{UInt64}
end

function AutoFwdDerivatives(fn, n; cs=min(16, n))
  r = DiffResults.HessianResult(zeros(n))
  cfg = ForwardDiff.HessianConfig(fn, r, zeros(n), ForwardDiff.Chunk{cs}())
  AutoFwdDerivatives(fn, r, cfg, [hash([NaN])])
end

function _gradient(buf, D::AutoFwdDerivatives{F,R,C}, x) where{F,R,C} 
  D.hhash[] == hash(x) && return DiffResults.gradient(D.hess_res)
  ForwardDiff.gradient!(buf, D.fn, x)
end

function _hessian(buf, D::AutoFwdDerivatives{F,R,C}, x) where{F,R,C}
  D.hhash[] == hash(x) && return DiffResults.hessian(D.hess_res)
  ForwardDiff.hessian!(D.hess_res, D.fn, x, D.hess_cfg)
  D.hhash[1] = hash(x)
  buf .= DiffResults.hessian(D.hess_res)
end

struct AutoMixedDerivatives{W<:FoRWrapper} <: DerivativeSpec
  wrapper::W
end

function AutoMixedDerivatives(fn, n; cs=min(16, n))
  wrapper = FoRWrapper(fn, zeros(n), cs)
  AutoMixedDerivatives(wrapper)
end

function _gradient(buf, D::AutoMixedDerivatives{W}, x) where{W} 
  buf .= D.wrapper(:gradient, x)
end

function _hessian(buf, D::AutoMixedDerivatives{W}, x) where{W}
  buf .= D.wrapper(:hessian, x)
end

struct Derivatives{G,H} <: DerivativeSpec
  g::G
  h::H
end

_gradient(buf, D::Derivatives, x) = D.g(buf, x)
_hessian(buf, D::Derivatives, x)  = D.h(buf, x)

# a simple wrapper to avoid a struct:
struct ObjectiveGradient{D} <: Function
  derivatives::D 
end

function (og::ObjectiveGradient{D})(x, store) where{D}
  _gradient(store, og.derivatives, x)
end

