
struct DummyZero <: Function end
(dz::DummyZero)(x)    = zero(eltype(x))
(dz::DummyZero)(p, g) = fill!(g, zero(eltype(p)))

throw_error(args...) = throw(error("This method should never be reached!"))

function ipopt_nlsolve(fn, ini; box_lower=-1e22, box_upper=1e22, kwargs...)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  nconstr = arglen
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  con_l = zeros(length(ini))
  con_u = zeros(length(ini))
  # function for the constraints:
  constraints = Constraints(arglen, fn, con_l, con_u)
  # "constraint" jacobian:
  con_buf = zeros(arglen)
  jac_buf = zeros(arglen, arglen)
  conderivs  = AutoFwdConDerivatives(constraints)
  constr_jac = ConstraintJacobian(conderivs, jac_buf, con_buf)
  # Create the final problem:
  prob = CreateIpoptProblem(arglen,  box_l, box_u,        # arg count and box
                            nconstr, con_l, con_u,        # constr. count and box
                            arglen^2,                     # size of constr. jacobian
                            div(arglen*(arglen+1), 2),    # size of hessian ltri
                            DummyZero(),                  # objective
                            constraints,                  # constraint function
                            DummyZero(),                  # objective gradient
                            constr_jac,                   # constraint jacobian
                            throw_error)                  # lagrangian hessian
  # Add all the options:
  ipopt_addoption!(prob, "hessian_approximation", "limited-memory")
  ipopt_addoption!(prob, "nlp_scaling_method",    "none")
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=Ipopt._STATUS_CODES[status], minimizer=copy(prob.x))
end

