
struct Lagrangian{O,C} <: Function
  obj_derivs::O
  con_derivs::C
  hesbuf::Matrix{Float64}
  hestmp::Matrix{Float64}
end

function (lag::Lagrangian{O,C})(p, r, c, o, l, v) where{O,C}
  isnothing(v) && return hes_structure!(r,c,length(p)) 
  if o != 0.0
    _hessian(lag.hesbuf, lag.obj_derivs, p) 
  else
    fill!(lhessbuf, 0.0)
  end
  nc = lag.con_derivs.cons.ncon # TODO (cg 2023/01/07 17:36): write a getter
  lv = view(l, 1:nc)
  if nc > 0 && !all(iszero, lv)
    _con_lagpart_hess(lag.hestmp, lv, lag.con_derivs, p)
    lag.hesbuf .+= lag.hestmp
  end
  linear_buffer_ltri_colmajor!(v, lag.hesbuf) 
  nothing
end

function ipopt_optimize(obj, ini, constraints=noconstraints(); wrap=true,
                        derivatives=AutoFwdDerivatives(obj, length(ini)),
                        conderivatives=AutoFwdConDerivatives(constraints),
                        box_lower=-1e22, box_upper=1e22, kwargs...)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  # function for the constraints:
  constr_buf = zeros(constraints.ncon)
  # objective gradient and constraint jacobian:
  objg    = ObjectiveGradient(derivatives)
  jac_buf = zeros(constraints.ncon, arglen)
  jac_nz  = arglen*constraints.ncon
  constr_jac = ConstraintJacobian(conderivatives, jac_buf, constr_buf)
  # Lagrangian hessian:
  (lhessbuf, hesstmp) = (zeros(arglen, arglen), zeros(arglen, arglen))
  lagrange_h = Lagrangian(derivatives, conderivatives, lhessbuf, hesstmp)
  # create the wrapped or un-wrapped objective:
  ls_obj = wrap ? WrappedObjective(obj) : obj
  # Create the final problem:
  prob = CreateIpoptProblem(arglen,  box_l, box_u,     # arg count and box
                            constraints.ncon,          # number of constraints
                            constraints.box_lower,     # constraints lower box
                            constraints.box_upper,     # constraints upper box
                            jac_nz,                    # size of constr. jacobian
                            div(arglen*(arglen+1), 2), # size of hessian ltri
                            ls_obj,                    # objective
                            constraints,               # constraint function
                            objg,                      # objective gradient
                            constr_jac,                # constraint jacobian
                            lagrange_h)                # lagrangian hessian
  # Add all the options:
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=Ipopt._STATUS_CODES[status], minimizer=copy(prob.x), minval=obj(prob.x))
end

