
struct WrappedObjective{F} <: Function
  fun::F
end

function (wo::WrappedObjective{F})(x) where{F}
  try
    return wo.fun(x)
  catch er
    (er isa InterruptException) && rethrow(er)
    return NaN
  end
end

prepare_bounds(n, arg::Float64) = fill(arg, n)
prepare_bounds(n, argv::Vector{Float64}) = argv

# Could also use dispatch for this, but not sure it would be faster? It would
# look nicer than this though.
function ipopt_addoption!(prob, opt, val)
  if val isa Symbol
    AddIpoptStrOption(prob, string(opt), string(val))
  elseif val isa AbstractString
    AddIpoptStrOption(prob, string(opt), string(val))
  elseif isinteger(val)
    AddIpoptIntOption(prob, string(opt), Int64(val))
  else
    AddIpoptNumOption(prob, string(opt), val)
  end
  nothing
end

# the structure function for the hessian.
function hes_structure!(rows, cols, len)
  idx = 1
  for row in 1:len
    for col in 1:row
      rows[idx] = row
      cols[idx] = col
      idx += 1
    end
  end
  nothing
end

# Note that ipopt wants the TRANSPOSE of the constraint Jacobian, so the j and k
# in these loops will look wrong. This function could probably be written more
# simply than this, but I think the readability here is worth it, and all the
# utility functions here are ugly anyway.
function jac_structure!(rows, cols, nargs, nconstr)
  linear_idx = 1
  for j in 1:nargs
    for k in 1:nconstr
      rows[linear_idx] = j
      cols[linear_idx] = k
      linear_idx += 1
    end
  end
  nothing
end

# basically "vec" for a lower triangular matrix.
function linear_buffer_ltri_colmajor!(lbuf, mbuf)
  sz = size(mbuf, 2)
  linear_idx = 1
  for col_idx in 1:sz
    for row_idx in 1:col_idx
      lbuf[linear_idx] = mbuf[row_idx,col_idx]
      linear_idx += 1
    end
  end
  nothing
end

#=
function jac_structure_sparse!(jac_sparsity::SparseMatrixCSC{Bool, Int64}, 
                               rows, cols)
  linear_idx = 1
  for j in 1:size(jac_sparsity, 1)
    for k in 1:size(jac_sparsity, 2)
      val = jac_sparsity[j,k]
      if val
        rows[linear_idx] = j
        cols[linear_idx] = k
        linear_idx += 1
      end
    end
  end
  nothing
end
=#


#=
function populate_sparse_jacobian!(v, jac_sparsity::SparseMatrixCSC{Bool, Int64}, 
                                   buf::SparseMatrixCSC{Float64, Int64})
  linear_idx = 1
  for j in 1:size(jac_sparsity, 1)
    for k in 1:size(jac_sparsity, 2)
      if jac_sparsity[j,k]
        v[linear_idx] = buf[j,k]
        linear_idx   += 1
      end
    end
  end
  nothing
end
=#

